@@ -22,8 +22,9 @@ import logging
 import math
 import urllib
 
+from eventlet import tpool
 from oslo_config import cfg
-from oslo_utils import encodeutils
+from oslo_utils import eventletutils
 from oslo_utils import units
 
 from glance_store import capabilities
@@ -132,7 +133,7 @@ Related options:
     * rbd_store_user
 
 """),
-    cfg.IntOpt('rados_connect_timeout', default=0,
+    cfg.IntOpt('rados_connect_timeout', default=-1,
                help="""
 Timeout value for connecting to Ceph cluster.
 
@@ -140,8 +141,8 @@ This configuration option takes in the timeout value in seconds used
 when connecting to the Ceph cluster i.e. it sets the time to wait for
 glance-api before closing the connection. This prevents glance-api
 hangups during the connection to RBD. If the value for this option
-is set to less than or equal to 0, no timeout is set and the default
-librados value is used.
+is set to less than 0, no timeout is set and the default librados value
+is used.
 
 Possible Values:
     * Any integer value
@@ -258,7 +259,8 @@ class ImageIterator(object):
                                            rados_id=self.user) as conn:
                 with conn.open_ioctx(self.pool) as ioctx:
                     with rbd.Image(ioctx, self.name,
-                                   snapshot=self.snapshot) as image:
+                                   snapshot=self.snapshot,
+                                   read_only=True) as image:
                         size = image.size()
                         bytes_left = size
                         while bytes_left > 0:
@@ -280,15 +282,37 @@ class Store(driver.Store):
 
     EXAMPLE_URL = "rbd://<FSID>/<POOL>/<IMAGE>/<SNAP>"
 
+    def __init__(self, *args, **kargs):
+        super(Store, self).__init__(*args, **kargs)
+        if self.backend_group:
+            self.store_conf = driver.BackendGroupConfiguration(
+                self.OPTIONS, self.backend_group, conf=self.conf)
+        else:
+            self.store_conf = self.conf.glance_store
+
     def get_schemes(self):
         return ('rbd',)
 
+    def RBDProxy(self):
+        if eventletutils.is_monkey_patched('thread'):
+            return tpool.Proxy(rbd.RBD())
+        else:
+            return rbd.RBD()
+
     @contextlib.contextmanager
     def get_connection(self, conffile, rados_id):
         client = rados.Rados(conffile=conffile, rados_id=rados_id)
 
+        timeout = self.store_conf.rados_connect_timeout
+
+        if timeout >= 0:
+            t = str(timeout)
+            client.conf_set('rados_osd_op_timeout', t)
+            client.conf_set('rados_mon_op_timeout', t)
+            client.conf_set('client_mount_timeout', t)
+
         try:
-            client.connect(timeout=self.connect_timeout)
+            client.connect()
         except (rados.Error, rados.ObjectNotFound) as e:
             if self.backend_group and len(self.conf.enabled_backends) > 1:
                 reason = _("Error in store configuration: %s") % e
@@ -311,27 +335,19 @@ class Store(driver.Store):
         this method. If the store was not able to successfully configure
         itself, it should raise `exceptions.BadStoreConfiguration`
         """
+        if rbd is None or rados is None:
+            reason = _("The required libraries(rbd and rados) are not "
+                       "available")
+            LOG.error(reason)
+            raise exceptions.BadStoreConfiguration(store_name='rbd',
+                                                   reason=reason)
+
         try:
-            if self.backend_group:
-                chunk = getattr(self.conf,
-                                self.backend_group).rbd_store_chunk_size
-                pool = getattr(self.conf, self.backend_group).rbd_store_pool
-                user = getattr(self.conf, self.backend_group).rbd_store_user
-                conf_file = getattr(self.conf,
-                                    self.backend_group).rbd_store_ceph_conf
-                connect_timeout = getattr(
-                    self.conf, self.backend_group).rados_connect_timeout
-                thin_provisioning = getattr(self.conf,
-                                            self.backend_group).\
-                    rbd_thin_provisioning
-            else:
-                chunk = self.conf.glance_store.rbd_store_chunk_size
-                pool = self.conf.glance_store.rbd_store_pool
-                user = self.conf.glance_store.rbd_store_user
-                conf_file = self.conf.glance_store.rbd_store_ceph_conf
-                connect_timeout = self.conf.glance_store.rados_connect_timeout
-                thin_provisioning = \
-                    self.conf.glance_store.rbd_thin_provisioning
+            chunk = self.store_conf.rbd_store_chunk_size
+            pool = self.store_conf.rbd_store_pool
+            user = self.store_conf.rbd_store_user
+            conf_file = self.store_conf.rbd_store_ceph_conf
+            thin_provisioning = self.store_conf.rbd_thin_provisioning
 
             self.thin_provisioning = thin_provisioning
             self.chunk_size = chunk * units.Mi
@@ -343,7 +359,6 @@ class Store(driver.Store):
             self.pool = str(pool)
             self.user = str(user)
             self.conf_file = str(conf_file)
-            self.connect_timeout = connect_timeout
         except cfg.ConfigFileValueError as e:
             reason = _("Error in store configuration: %s") % e
             LOG.error(reason)
@@ -359,7 +374,7 @@ class Store(driver.Store):
         with self.get_connection(conffile=self.conf_file,
                                  rados_id=self.user) as conn:
             if hasattr(conn, 'get_fsid'):
-                fsid = encodeutils.safe_decode(conn.get_fsid())
+                fsid = conn.get_fsid()
 
         if fsid and self.pool:
             # ensure nothing contains / or any other url-unsafe character
@@ -402,7 +417,8 @@ class Store(driver.Store):
             with conn.open_ioctx(target_pool) as ioctx:
                 try:
                     with rbd.Image(ioctx, loc.image,
-                                   snapshot=loc.snapshot) as image:
+                                   snapshot=loc.snapshot,
+                                   read_only=True) as image:
                         img_info = image.stat()
                         return img_info['size']
                 except rbd.ImageNotFound:
@@ -421,12 +437,12 @@ class Store(driver.Store):
 
         :returns: `glance_store.rbd.StoreLocation` object
         """
-        librbd = rbd.RBD()
         features = conn.conf_get('rbd_default_features')
         if ((features is None) or (int(features) == 0)):
             features = rbd.RBD_FEATURE_LAYERING
-        librbd.create(ioctx, image_name, size, order, old_format=False,
-                      features=int(features))
+        self.RBDProxy().create(ioctx, image_name, size, order,
+                               old_format=False,
+                               features=int(features))
         return StoreLocation({
             'fsid': fsid,
             'pool': self.pool,
@@ -434,14 +450,6 @@ class Store(driver.Store):
             'snapshot': DEFAULT_SNAPNAME,
         }, self.conf)
 
-    def _snapshot_has_external_reference(self, image, snapshot_name):
-        """Returns True if snapshot has external reference else False.
-        """
-        image.set_snap(snapshot_name)
-        has_references = bool(image.list_children())
-        image.set_snap(None)
-        return has_references
-
     def _delete_image(self, target_pool, image_name,
                       snapshot_name=None, context=None):
         """
@@ -461,14 +469,6 @@ class Store(driver.Store):
                     if snapshot_name is not None:
                         with rbd.Image(ioctx, image_name) as image:
                             try:
-                                # NOTE(abhishekk): Check whether snapshot
-                                # has any external references
-                                if self._snapshot_has_external_reference(
-                                        image, snapshot_name):
-                                    raise rbd.ImageBusy(
-                                        "Image snapshot has external "
-                                        "references.")
-
                                 self._unprotect_snapshot(image, snapshot_name)
                                 image.remove_snap(snapshot_name)
                             except rbd.ImageNotFound as exc:
@@ -486,12 +486,21 @@ class Store(driver.Store):
                                 raise exceptions.InUseByStore()
 
                     # Then delete image.
-                    rbd.RBD().remove(ioctx, image_name)
+                    self.RBDProxy().remove(ioctx, image_name)
                 except rbd.ImageHasSnapshots:
-                    log_msg = (_LW("Remove image %(img_name)s failed. "
-                                   "It has snapshot(s) left.") %
+                    log_msg = (_LW("Unable to remove image %(img_name)s: it "
+                                   "has snapshot(s) left; trashing instead") %
                                {'img_name': image_name})
                     LOG.warning(log_msg)
+                    with rbd.Image(ioctx, image_name) as image:
+                        try:
+                            rbd.RBD().trash_move(ioctx, image_name)
+                            LOG.debug('Moved %s to trash', image_name)
+                        except rbd.ImageBusy:
+                            LOG.warning(_('Unable to move in-use image to '
+                                          'trash'))
+                            raise exceptions.InUseByStore()
+                        return
                     raise exceptions.HasSnapshot()
                 except rbd.ImageBusy:
                     log_msg = (_LW("Remove image %(img_name)s failed. "
@@ -516,12 +525,12 @@ class Store(driver.Store):
         """Handle the rbd resize when needed."""
         if image_size != 0 or self.size >= bytes_written + chunk_length:
             return self.size
-        new_size = self.size + self.resize_amount
-        LOG.debug("resizing image to %s KiB" % (new_size / units.Ki))
-        image.resize(new_size)
         # Note(jokke): We double how much we grow the image each time
         # up to 8gigs to avoid resizing for each write on bigger images
         self.resize_amount = min(self.resize_amount * 2, 8 * units.Gi)
+        new_size = self.size + self.resize_amount
+        LOG.debug("resizing image to %s KiB" % (new_size / units.Ki))
+        image.resize(new_size)
         return new_size
 
     @driver.back_compat_add
@@ -553,18 +562,7 @@ class Store(driver.Store):
                                  rados_id=self.user) as conn:
             fsid = None
             if hasattr(conn, 'get_fsid'):
-                # Librados's get_fsid is represented as binary
-                # in py3 instead of str as it is in py2.
-                # This is causing problems with ceph.
-                # Decode binary to str fixes these issues.
-                # Fix with encodeutils.safe_decode CAN BE REMOVED
-                # after librados's fix will be stable.
-                #
-                # More information:
-                # https://bugs.launchpad.net/glance-store/+bug/1816721
-                # https://bugs.launchpad.net/cinder/+bug/1816468
-                # https://tracker.ceph.com/issues/38381
-                fsid = encodeutils.safe_decode(conn.get_fsid())
+                fsid = conn.get_fsid()
             with conn.open_ioctx(self.pool) as ioctx:
                 order = int(math.log(self.WRITE_CHUNKSIZE, 2))
                 LOG.debug('creating image %s with order %d and size %d',
@@ -597,6 +595,14 @@ class Store(driver.Store):
                                                               bytes_written,
                                                               chunk_length)
                             bytes_written += chunk_length
+                            # Check if writing this chunk exceeds the
+                            # image_size
+                            if image_size != 0 and bytes_written > image_size:
+                                raise exceptions.Invalid(
+                                    _("Size exceeds: expected %(expected)d "
+                                      "bytes, got %(actual)d bytes") %
+                                    {'expected': image_size,
+                                     'actual': bytes_written})
                             if not (self.thin_provisioning and not any(chunk)):
                                 image.write(chunk, offset)
                             offset += chunk_length
@@ -627,6 +633,23 @@ class Store(driver.Store):
                         pass
 
                     raise exceptions.StorageFull(message=log_msg)
+                except rbd.IncompleteWriteError as exc:
+                    log_msg = (_LE("Failed to store image %(img_name)s "
+                                   "image data exceeded the expected size. "
+                                   "Store Exception %(store_exc)s") %
+                               {'img_name': image_name,
+                                'store_exc': exc})
+                    LOG.error(log_msg)
+
+                    # Delete image if one was created
+                    try:
+                        target_pool = loc.pool or self.pool
+                        self._delete_image(target_pool, loc.image,
+                                           loc.snapshot)
+                    except exceptions.NotFound:
+                        pass
+
+                    raise exceptions.Invalid(message=log_msg)
                 except Exception as exc:
                     log_msg = (_LE("Failed to store image %(img_name)s "
                                    "Store Exception %(store_exc)s") %
@@ -644,9 +667,20 @@ class Store(driver.Store):
 
                     raise exc
 
-        # Make sure we send back the image size whether provided or inferred.
-        if image_size == 0:
-            image_size = bytes_written
+        # If actual bytes written are less than image_size
+        # (when image_size != 0), update accordingly
+        if image_size != 0 and bytes_written != image_size:
+            # Delete image if one was created
+            try:
+                self._delete_image(loc.pool or self.pool, loc.image,
+                                   loc.snapshot)
+            except exceptions.NotFound:
+                pass
+
+            raise exceptions.Invalid(_(
+                "Size mismatch: expected %(expected)d "
+                "bytes, got %(actual)d bytes") % {
+                'expected': image_size, 'actual': bytes_written})
 
         # Add store backend information to location metadata
         metadata = {}
@@ -654,7 +688,7 @@ class Store(driver.Store):
             metadata['store'] = self.backend_group
 
         return (loc.get_uri(),
-                image_size,
+                bytes_written,
                 checksum.hexdigest(),
                 os_hash_value.hexdigest(),
                 metadata)
